/*-
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@jp.freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$FreeBSD$
 */

#include <sys/file.h>

#include <err.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <libpower.h>
#include <dev/power/powervar.h>

static void
print_syspm_info(struct power_syspm_info *infp)
{

	printf("AC Line status: ");
	if (infp->acline == POWER_ACLINE_STATUS_UNKNOWN)
		printf("unknown");
	else if (infp->acline > 1)
		printf("invalid value (0x%x)", infp->acline);
	else {
		char *messages[] = { "off-line", "on-line" };
		printf("%s", messages[infp->acline]);
	}
	printf("\n");
	printf("Battery status: ");
	if (infp->batt_stat == POWER_BATT_STATUS_UNKNOWN)
		printf("unknown");
	else if (infp->batt_stat > 3)
			printf("invalid value (0x%x)", infp->batt_stat);
	else {
		char *messages[] = { "high", "low", "critical", "charging" };
		printf("%s", messages[infp->batt_stat]);
	}
	printf("\n");
	printf("Remaining battery life: ");
	if (infp->batt_life == POWER_BATT_LIFE_UNKNOWN)
		printf("unknown\n");
	else if (infp->batt_life <= 100)
		printf("%d%%\n", infp->batt_life);
	else
		printf("invalid value (0x%x)\n", infp->batt_life);
	printf("Remaining battery time: ");
	if (infp->batt_time == POWER_BATT_TIME_UNKNOWN)
		printf("unknown\n");
	else {
		int t, h, m, s;

		t = infp->batt_time;
		s = t % 60;
		t /= 60;
		m = t % 60;
		t /= 60;
		h = t;
		printf("%2d:%02d:%02d\n", h, m, s);
	}

	printf("Number of batteries: ");
	if (infp->batteries == POWER_BATTERIES_UNKNOWN)
		printf("unknown\n");
	else {
		printf("%d\n", infp->batteries);
	}
}

static void
print_batt_info(int unit, struct power_batt_info *battp)
{

	printf("Battery %d:\n", unit);
	printf("\tBattery status: ");
	if (battp->batt_flag != POWER_BATT_FLAG_UNKNOWN &&
	    (battp->batt_flag & POWER_BATT_FLAG_NOT_PRESENT)) {
		printf("not present\n");
		return;
	}
	if (battp->batt_stat == POWER_BATT_STATUS_UNKNOWN)
		printf("unknown\n");
	else if (battp->batt_stat > 3)
		printf("invalid value (0x%x)\n",
		    battp->batt_stat);
	else {
		char *messages[] = { "high",
				     "low",
				     "critical",
				     "charging" };
		printf("%s\n", messages[battp->batt_stat]);
	}
	printf("\tRemaining battery life: ");
	if (battp->batt_life == POWER_BATT_TIME_UNKNOWN)
		printf("unknown\n");
	else if (battp->batt_life <= 100)
		printf("%d%%\n", battp->batt_life);
	else
		printf("invalid value (0x%x)\n", battp->batt_life);
	printf("\tRemaining battery time: ");
	if (battp->batt_time == POWER_BATT_TIME_UNKNOWN)
		printf("unknown\n");
	else {
		int	 t, h, m, s;

		t = battp->batt_time;
		s = t % 60;
		t /= 60;
		m = t % 60;
		t /= 60;
		h = t;
		printf("%2d:%02d:%02d\n", h, m, s);
	}
}

static int
print_pm_info()
{
	int	 i;
	struct	 power_syspm_info info;
	struct	 power_batt_info batt;

	if (power_get_syspm_info(&info) == -1) {
		return (-1);
	}

	print_syspm_info(&info);
	if (info.batteries == POWER_BATTERIES_UNKNOWN) {
		return (0);
	}

	for (i = 0; i < info.batteries; i++) {
		bzero(&batt, sizeof(batt));
		if (power_get_batt_info(i, &batt) == -1)
			continue;
		print_batt_info(i, &batt);
	}
	return (0);
}

int 
main(int argc, char *argv[])
{
	char	 c;

	while ((c = getopt(argc, argv, "Ssz")) != -1) {
		switch (c) {
		case 'S':
			return (power_standby());

		case 's':
			return (power_suspend());

		case 'z':
			return (power_hibernate());

		default:
			argc -= optind;
			argv += optind;
                        break;
		}
	}

	return (print_pm_info());
}

